{{

  Propeller UDP PacketLink Serial Bridge Server-Only Mini Version

  This implements a packet-based serial-ethernet bridge, packets being delimited
  by a fixed terminating character, with packets not being sent until complete
  and sent immediately when the terminator arrives.  There is no re-ordering or
  replacement of missing or out of sequence packets.

  This very simple UDP implementation only supports one connection.  As a server
  it will always send to the last client from which a message was received.
  As as server we do not do ARP and do not concern ourselves too much whether
  the client is still online, as servers will generally mostly be client-driven
  receivers anyway.

  In the interest of minimal size this version does not act as a client.  A client
  must send the first data to this server to initiate communication.

  An additional 40 longs can be saved by commenting out handle_ping, although this
  could make troubleshooting difficult.

  OBJECT MODS:
  driver_enc28j60 has been modified to return partial packets
  fd_serial has been modified to add txcheck to avoid buffer overruns
  
}}

CON
  _clkmode = xtal1+pll16x
  _xinfreq = 5_000_000

  ser_buf_max = ser#bufsiz 'effective maximum size of ser to eth packets
  eth_pkt_max = 128 'nic#maxframe
  {
  ser to eth packets are limited by the ser_rxbuf[ser_buf_max] which is
  used to collect characters until a packet is ready, while other ethernet
  stuff is happening in packet[].  It is ALSO limited by the size of
  packet[], which must be 42 longer than the longest serial packet to
  be handled.  Normally this is nic#maxframe, the maximum size of an
  ethernet packet.

  eth to ser packets are limited by nic#maxframe but also by the buffer
  size specified in fds_udp_rr; if a frame arrives with more data than
  the serial buffer can take it will overrun and data will be lost.
  
  If these are the only protocols implemented and long serial packets do not
  need to be handled, eth_pkt_max can be reduced to 42 + longest desired
  serial packet or 64, whichever is larger, to save space for packet[].
  eth_pkt_max needs to be at least 80 or so to handle most implementations
  of PING (with 32 bytes dummy data).
  }
DAT

  'SETUPS
  '
  'IP addresses must be long-aligned.  They are treated as Propeller Longs
  'in cases where the reverse byte ordering does not affect results.
  '
                long 
  ip_addr       byte    192, 168,   1,  80              ' device's ip address

  udp_port      word    48675                           ' port the client uses to call in either direction

  nic_cs        byte    23    'SVT1 ENC28J60 pins
  nic_sck       byte    20
  nic_si        byte    21
  nic_so        byte    22
  nic_int       byte    -1
  nic_xtalout   byte    -1
  mac_addr      byte    $02, $01, $02, $03, $03, $05    ' device mac address, must be unique

  ser_tx        byte    16
  ser_rx        byte    17
  ser_mode      byte    0
  ser_baud      long    57600 '9600

  ser_eom       byte    $0D  'ser to eth packet EOM character


OBJ

  nic : "driver_enc28j60_udp"   'enc28j60 with get_frame modified to return partial frame
  ser : "fds_udp_rr"            'fullduplexserial + rr004 sizeable buffer + txcheck

VAR

  byte ser_rxbuf[ser_buf_max]
  word ser_rxbuf_ptr
  
  long arp_cnt        'base cnt for all arp timeouts
  word arp_down_sec   'down seconds count
  byte arp_down       '0=online 1-arp_retries=retrying >=arp_retries=in down interval
  byte arp_set        'clear until _any_ mac resolution occurs and we can at least try to send

  byte packet[eth_pkt_max]
  long other_mac_addr[2]
  byte other_mac_known
  long other_ip_addr
  long other_port_cmp   'NB no un-_cmp version since this
                        'is taken pre-reversed from packet

  byte pkt_count      'packet count

  long ip_pkt_id
  
  word udp_port_cmp      'ports must be byte reversed for direct
  '

PUB start | rxch

  ser.start(ser_rx, ser_tx, ser_mode, ser_baud)

  'create byte reversed comparison values for UDP ports
  '
  udp_port_cmp := udp_port >> 8 + (udp_port & $FF) << 8 

  'NB int and xtalout must be sign extended so bytes become long -1 to disable
  '
  nic.start(nic_cs, nic_sck, nic_si, nic_so, ~nic_int, ~nic_xtalout, @mac_addr)

  'ser.tx("%")

  waitcnt(clkfreq / 10 + cnt) 'give ethernet time to initialize
  
  'main loop
  '
  nic.banksel(nic#EPKTCNT)                                                      ' select packet count bank
  repeat
    pkt_count := nic.rd_cntlreg(nic#EPKTCNT)
    if pkt_count > 0
      service_packet                                                            ' handle packet
      nic.banksel(nic#EPKTCNT)                                                  ' re-select the packet count bank
    '
    'accept serial data and see if we have a packet to send
    '
    repeat while (rxch := ser.rxcheck) => 0
      ser_rxbuf[ser_rxbuf_ptr++] := rxch
      if ser_rxbuf_ptr == ser_buf_max or rxch == ser_eom
        'buffer full OR packet complete
        send_UDP(@ser_rxbuf, ser_rxbuf_ptr - 1)
        ser_rxbuf_ptr := 0
        nic.banksel(nic#EPKTCNT)  
     
PRI service_packet | udplen, i

  'lets process this frame
  nic.get_frame(@packet, eth_pkt_max)

  'check for arp packet
  if pkt_cmp(enetpacketType0, 2, $608) 'ARP $0806
    'Next test is only satisified if
    'arp_hwtype=$0001, arp_prtype=$8000, arp_hwlen=6, arp_prlen=4:
    if pkt_cmp(arp_hwtype,4,$80100) AND pkt_cmp(arp_hwlen, 2, $406)
      if pkt_cmp(arp_tipaddr, 4, long[@ip_addr]) 'our IP
        case packet[constant(arp_op + 1)]
          $01 : handle_arp

  'check for IPV4 packet   
  elseif pkt_cmp(enetpacketType0, 2, $8) 'IPV4 $0800
    if pkt_cmp(ip_destaddr, 4, long[@ip_addr])   'our IP
      case packet[ip_proto]
        PROT_ICMP : handle_ping
        PROT_TCP :  '++count_tcp
        PROT_UDP :  handle_udp
           
'Answer an arp request from someone else on the network...        

PRI handle_arp | i

  nic.start_frame

  ' destination mac address
  repeat i from 0 to 5
    nic.wr_frame(packet[enetpacketSrc0 + i])

  ' source mac address
  repeat i from 0 to 5
    nic.wr_frame(mac_addr[i])

  nic.wr_frame($08)             ' arp packet
  nic.wr_frame($06)

  nic.wr_frame($00)             ' 10mb ethernet
  nic.wr_frame($01)

  nic.wr_frame($08)             ' ip proto
  nic.wr_frame($00)

  nic.wr_frame($06)             ' mac addr len
  nic.wr_frame($04)             ' proto addr len

  nic.wr_frame($00)             ' arp reply
  nic.wr_frame($02)

  ' write ethernet module mac address
  repeat i from 0 to 5
    nic.wr_frame(mac_addr[i])

  ' write ethernet module ip address
  repeat i from 0 to 3
    nic.wr_frame(ip_addr[i])

  ' write remote mac address
  repeat i from 0 to 5
    nic.wr_frame(packet[enetpacketSrc0 + i])

  ' write remote ip address
  repeat i from 0 to 3
    nic.wr_frame(packet[arp_sipaddr + i])

  return nic.send_frame


PRI handle_ping | swapbuf[2], cklen, cksum, pktlen, i
  '
  pktlen := packet[ip_pktlen] << 8 + packet[constant(ip_pktlen+1)] + ip_vers_len
  '
  'swap mac addreses
  '
  bytemove(@swapbuf, @packet+enetpacketsrc0, 6)
  bytemove(@packet+enetpacketsrc0,  @packet+enetpacketdest0, 6)
  bytemove(@packet+enetpacketdest0, @swapbuf, 6)
  '
  ' set ip id
  '
  ++ip_pkt_id
  packet[ip_id] := ip_pkt_id >> 8
  packet[constant(ip_id + 1)] := ip_pkt_id
  '
  'swap IP addresses
  '
  bytemove(@swapbuf, @packet+ip_srcaddr, 4)
  bytemove(@packet+ip_srcaddr, @packet+ip_destaddr, 4)
  bytemove(@packet+ip_destaddr, @swapbuf, 4)
  '
  ' calc ip header checksum
  '
  ip_checksum
  '
  ' set ICMP type to echo reply
  '
  packet[icmp_type] := 0
  '
  'zero out original ICMP checksum fields
  '
  packet[icmp_cksum] := 0
  packet[constant(icmp_cksum+1)] := 0
  '
  'build new checksum
  '
  cksum := calc_chksum(@packet + ip_data, pktlen - ip_data)
  packet[icmp_cksum] := cksum >> 8
  packet[constant(icmp_cksum+1)] := cksum
  '
  'send the packet
  '
  nic.start_frame  
  nic.wr_block(@packet, pktlen)
  nic.send_frame

PRI handle_udp | udplen, i
  '
  'we're a server, only listen to our host port
  if not pkt_cmp(udp_destport, 2, udp_port_cmp)
    return
  '
  'and remember who to reply to
  bytemove(@other_ip_addr, @packet[ip_srcaddr],4)
  bytemove(@other_mac_addr, @packet[enetpacketsrc0], 6)
  bytemove(@other_port_cmp, @packet[udp_srcport], 2)
  '
  other_mac_known~~     
  
  udplen := packet[udp_len]<<8 + packet[udp_len+1]
  if udplen > eth_pkt_max - 34
    udplen := eth_pkt_max - 34
  i := udp_data
  repeat while i < udplen+constant(udp_data-8) and ser.txcheck
    ser.tx(packet[i++])

PRI set_UDP_len(udplen) | iplen
  '
  iplen := udplen + 20  'add IP to UDP
  '
  packet[ip_pktlen] := iplen >> 8
  packet[constant(ip_pktlen+1)] := iplen
  '
  packet[udp_len] := udplen >> 8
  packet[constant(udp_len+1)] := udplen  


PRI send_UDP(newdata, newlen)  | udplen, pktlen, i 
  '
  if not other_mac_known
    return
  '
  'ethernet header
  '
  bytemove(@packet+enetpacketdest0, @other_mac_addr, 6)
  bytemove(@packet+enetpacketsrc0,  @mac_addr, 6)
  packet[enetpacketType0] := $08
  packet[constant(enetpacketType0 + 1)] := $00
  '
  'ip header
  '
  packet[ip_vers_len] := $45
  packet[ip_tos] := $00
  ++ip_pkt_id
  packet[ip_id] := ip_pkt_id >> 8                                            ' Used for fragmentation
  packet[constant(ip_id + 1)] := ip_pkt_id
  packet[ip_frag_offset] := $40                                              ' Don't fragment
  packet[constant(ip_frag_offset + 1)] := 0
  packet[ip_ttl] := $80                                                      ' TTL = 128
  packet[ip_proto] := PROT_UDP
  '
  bytemove(@packet[ip_srcaddr], @ip_addr, 4)
  '
  'udp header
  '
  'as server we use dest ip and other-eph port given to us by client
  '
  bytemove(@packet[ip_destaddr], @other_ip_addr, 4)
  bytemove(@packet[udp_srcport], @udp_port_cmp, 2)
  bytemove(@packet[udp_destport], @other_port_cmp, 2)
   
  'data
  '
  if newlen > eth_pkt_max - 42
    newlen := eth_pkt_max - 42
  bytemove(@packet[udp_data], newdata, newlen)
  '
  'length
  '
  udplen := newlen + 8  'include UDP header
  set_udp_len(udplen)
  '
  'checksums
  '
  ip_checksum
  tcp_checksum(udp_cksum, udplen+20)
  '  
  'minimum packet length 60 bytes 
  '
  pktlen := udplen + ip_data  'include IP and ethernet header
  if pktlen < 60  'minimum packet length
    pktlen := 60
  '
  'send the packet
  '
  nic.start_frame  
  nic.wr_block(@packet, pktlen) '
  nic.send_frame


'' Utilities
'
' Compare 2 to 4 non-aligned bytes in packet
' to long aligned byte-reversed val
'
pri pkt_cmp(bufloc, len, val) | buf
  buf := 0
  bytemove(@buf, @packet[bufloc], len)
  return buf == val
'
' Calculate and insert the IP checksum
'
PRI ip_checksum | cklen, cksum
  '
  packet[ip_hdr_cksum] := $00
  packet[constant(ip_hdr_cksum + 1)] := $00
  cklen := (packet[ip_vers_len] & $0F) * 4
  cksum := calc_chksum(@packet[ip_vers_len], cklen)  
  packet[ip_hdr_cksum] := cksum >> 8
  packet[constant(ip_hdr_cksum + 1)] := cksum
  
PRI tcp_checksum(ck_ptr, ck_len) | hdr_chksum, i
  '
  ' calc TCP or UDP checksum with pseudo-header
  '
  packet[ck_ptr] := $00
  packet[ck_ptr + 1] := $00
  '
  hdr_chksum := nic.chksum_add(@packet[ip_srcaddr], 8)
  hdr_chksum += packet[ip_proto]
  i := ck_len - ((packet[ip_vers_len] & $0F) * 4)
  hdr_chksum += i
  hdr_chksum += nic.chksum_add(@packet[TCP_srcport], i)
  hdr_chksum := calc_chksumfinal(hdr_chksum)
  '
  packet[ck_ptr] := hdr_chksum >> 8
  packet[ck_ptr + 1] := hdr_chksum

PRI calc_chksum(ptr, hdrlen) : chksum
  ' Calculates IP checksums
  ' packet = pointer to IP packet
  ' returns: chksum
  ' http://www.geocities.com/SiliconValley/2072/bit33.txt
  'chksum := calc_chksumhalf(packet, hdrlen)
  chksum := nic.chksum_add(ptr, hdrlen)
  chksum := calc_chksumfinal(chksum)

PRI calc_chksumfinal(chksumin) : chksum
  ' Performs the final part of checksums
  chksum := (chksumin >> 16) + (chksumin & $FFFF)
  chksum := (!chksum) & $FFFF
  
CON
  '******************************************************************
  '*      TCP Flags
  '******************************************************************
  TCP_FIN = 1
  TCP_SYN = 2
  TCP_RST = 4
  TCP_PSH = 8
  TCP_ACK = 16
  TCP_URG = 32
  TCP_ECE = 64
  TCP_CWR = 128
  '******************************************************************
  '*      Ethernet Header Layout
  '******************************************************************                
  enetpacketDest0 = $00  'destination mac address
  enetpacketDest1 = $01
  enetpacketDest2 = $02
  enetpacketDest3 = $03
  enetpacketDest4 = $04
  enetpacketDest5 = $05
  enetpacketSrc0 = $06  'source mac address
  enetpacketSrc1 = $07
  enetpacketSrc2 = $08
  enetpacketSrc3 = $09
  enetpacketSrc4 = $0A
  enetpacketSrc5 = $0B
  enetpacketType0 = $0C  'type/length field
  enetpacketType1 = $0D
  enetpacketData = $0E  'IP data area begins here
  '******************************************************************
  '*      ARP Layout
  '******************************************************************
  arp_hwtype = $0E
  arp_prtype = $10
  arp_hwlen = $12
  arp_prlen = $13
  arp_op = $14
  arp_shaddr = $16   'arp source mac address
  arp_sipaddr = $1C   'arp source ip address
  arp_thaddr = $20   'arp target mac address
  arp_tipaddr = $26   'arp target ip address
  '******************************************************************
  '*      IP Header Layout
  '******************************************************************
  ip_vers_len = $0E       'IP version and header length 1a19
  ip_tos = $0F    'IP type of service
  ip_pktlen = $10 'packet length
  ip_id = $12     'datagram id
  ip_frag_offset = $14    'fragment offset
  ip_ttl = $16    'time to live
  ip_proto = $17  'protocol (ICMP=1, TCP=6, UDP=11)
  ip_hdr_cksum = $18      'header checksum 1a23
  ip_srcaddr = $1A        'IP address of source
  ip_destaddr = $1E       'IP addess of destination
  ip_data = $22   'IP data area
  '******************************************************************
  '*      TCP Header Layout
  '******************************************************************
  TCP_srcport = $22       'TCP source port
  TCP_destport = $24      'TCP destination port
  TCP_seqnum = $26        'sequence number
  TCP_acknum = $2A        'acknowledgement number
  TCP_hdrlen = $2E        '4-bit header len (upper 4 bits)
  TCP_hdrflags = $2F      'TCP flags
  TCP_window = $30        'window size
  TCP_cksum = $32 'TCP checksum
  TCP_urgentptr = $34     'urgent pointer
  TCP_data = $36 'option/data
  '******************************************************************
  '*      IP Protocol Types
  '******************************************************************
  PROT_ICMP = $01
  PROT_TCP = $06
  PROT_UDP = $11
  '******************************************************************
  '*      ICMP Header
  '******************************************************************
  ICMP_type = ip_data
  ICMP_code = ICMP_type+1
  ICMP_cksum = ICMP_code+1
  ICMP_id = ICMP_cksum+2
  ICMP_seqnum = ICMP_id+2
  ICMP_data = ICMP_seqnum+2
  '******************************************************************
  '*      UDP Header
  '******************************************************************
  UDP_srcport = ip_data
  UDP_destport = UDP_srcport+2
  UDP_len = UDP_destport+2
  UDP_cksum = UDP_len+2
  UDP_data = UDP_cksum+2
  '******************************************************************
  '*      DHCP Message
  '******************************************************************
  DHCP_op = UDP_data
  DHCP_htype = DHCP_op+1
  DHCP_hlen = DHCP_htype+1
  DHCP_hops = DHCP_hlen+1
  DHCP_xid = DHCP_hops+1
  DHCP_secs = DHCP_xid+4
  DHCP_flags = DHCP_secs+2
  DHCP_ciaddr = DHCP_flags+2
  DHCP_yiaddr = DHCP_ciaddr+4
  DHCP_siaddr = DHCP_yiaddr+4
  DHCP_giaddr = DHCP_siaddr+4
  DHCP_chaddr = DHCP_giaddr+4
  DHCP_sname = DHCP_chaddr+16
  DHCP_file = DHCP_sname+64
  DHCP_options = DHCP_file+128
  DHCP_message_end = DHCP_options+312    